#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <string.h>

#include <vector>
#include <map>
#include <iostream>
#include <future>
#include <thread>
#include <chrono>


class SpiTransfer
{
public:
  virtual int count() const = 0;
  virtual const struct spi_ioc_transfer * transfers() const = 0;
};

class OutputTransferFactory
{
public:
  virtual SpiTransfer * createOutputSpiTransfer(int i, uint8_t* buf) const = 0;
  virtual SpiTransfer * createConfigureSpiTransfer() const = 0;
};

enum { mcp23s17StartAddress = 0x40 };

enum class Mcp23s17Reg
  {
    IODIRA = 0x00,
    IODIRB = 0x01,
    IPOLA = 0x02,
    IPOLB = 0x03,
    GPINTENA = 0x04,
    GPINTENB = 0x05,
    DEFVALA = 0x06,
    DEFVALB = 0x07,
    INTCONA = 0x08,
    INTCONB = 0x09,
    IOCONA = 0x0A,
    IOCONB = 0x0B,
    IOCON = IOCONA,
    GPPUA = 0x0C,
    GPPUB = 0x0D,
    INTFA = 0x0E,
    INTFB = 0x0F,
    INTCAPA = 0x10,
    INTCAPB = 0x11,
    GPIOA = 0x12,
    GPIOB = 0x13,
    OLATA = 0x14,
    OLATB = 0x15
  };

static void mcp23s17FillSetRegisterTransfer(struct spi_ioc_transfer *t,
                                            uint8_t * buffer,
                                            int chipAddress,
                                            Mcp23s17Reg regno,
                                            int value)
{
  memset(t, 0, sizeof(*t));
  buffer[0] = (chipAddress<<1) + mcp23s17StartAddress;
  buffer[1] = uint8_t(regno);
  buffer[2] = value;
  t->tx_buf = (unsigned long)buffer;
  t->rx_buf = 0;
  t->len = 3;
  t->delay_usecs = 0;
  t->speed_hz = 0;
  t->bits_per_word = 8;
  t->cs_change = 0;
}



class Mcp23S17ConfigureTransfer : public SpiTransfer
{
  uint8_t buffer[3*3*8];
  struct spi_ioc_transfer m_transfers[3*8];
public:
  Mcp23S17ConfigureTransfer()
  {
    memset(m_transfers, 0, sizeof(m_transfers));

    // configure all 8 possible MCP23s17 on that chip-select:
    // 1. HAEN=1
    // 2. as outputs
    struct spi_ioc_transfer *t = m_transfers;
    for (auto i = 0; i < 8; ++i)
      {
        mcp23s17FillSetRegisterTransfer(t,
                                        &buffer[i*3*3+0],
                                        i,
                                        Mcp23s17Reg::IOCON,
                                        8); // HAEN=1: Enables the MCP23S17 address pins
        t->cs_change = 1;
        t++;

        mcp23s17FillSetRegisterTransfer(t,
                                        &buffer[i*3*3+3],
                                        i,
                                        Mcp23s17Reg::IODIRA,
                                        0); // All GPIOA are outputs
        t->cs_change = 1;
        t++;

        mcp23s17FillSetRegisterTransfer(t,
                                        &buffer[i*3*3+6],
                                        i,
                                        Mcp23s17Reg::IODIRB,
                                        0); // All GPIOB are outputs
        t->cs_change = 1;
        t++;
    }
    (t-1)->cs_change = 0;
  }

  int count() const
  {
    return sizeof(m_transfers)/sizeof(m_transfers[0]);
  }

  const struct spi_ioc_transfer * transfers() const
  {
    return m_transfers;
  }
};


class Mcp23S17OutputTransfer : public SpiTransfer
{
private:
  enum
    {
      ChannelsPerLayer = 64,
      BytesPerChip = 2,
      ChannelsPerPort = 8,
      LayerChipCount = ChannelsPerLayer / ChannelsPerPort / BytesPerChip
    };
  int       m_transferCount = 0;
  uint8_t   m_ledControlBuffer[LayerChipCount][2];
  uint8_t   m_columnBuffer[2*3];
  struct spi_ioc_transfer m_transfers[2*(LayerChipCount)+2];

public:
  Mcp23S17OutputTransfer (int column,
                          uint8_t *ledBuffer
                          )
  {
    m_transferCount = 0;
    memset(m_transfers, 0, sizeof(m_transfers));
    struct spi_ioc_transfer *tr = m_transfers;
    
    buildColumnSelectTransfer(tr,
                              &m_columnBuffer[0],
                              0);
    tr->cs_change = 1;
    tr++;
    m_transferCount++;

    for (auto i = 0; i < LayerChipCount; ++i)
      {
        {
          m_ledControlBuffer[i][0] = (i<<1) + mcp23s17StartAddress;
          m_ledControlBuffer[i][1] = int(Mcp23s17Reg::GPIOA);

          tr->tx_buf = (unsigned long)m_ledControlBuffer[i];
          tr->rx_buf = 0;
          tr->len = 2;
          tr->delay_usecs = 0;
          tr->speed_hz = 0;
          tr->bits_per_word = 8;
          tr->cs_change = 0; // -> hold CS low
          tr++;
          m_transferCount++;
        }
        {
          tr->tx_buf = (unsigned long)&ledBuffer[2*i];
          tr->rx_buf = 0;
          tr->len = 2;
          tr->delay_usecs = 0;
          tr->speed_hz = 0;
          tr->bits_per_word = 8;
          tr->cs_change = 1; // -> CS high
          tr++;
          m_transferCount++;
        }
      }
    buildColumnSelectTransfer(tr,
                              &m_columnBuffer[3],
                              (1<<column));
    tr->cs_change = 0;
    tr++;
    m_transferCount++;
    if (tr != &m_transfers[sizeof(m_transfers)/sizeof(*m_transfers)])
      throw std::runtime_error("invalid transfer count");
    std::cout << "trcount:" << m_transferCount << std::endl;
  }

  int count() const
  {
    return m_transferCount;
  }

  const struct spi_ioc_transfer * transfers() const
  {
    return m_transfers;
  }

private:
  void buildColumnSelectTransfer(struct spi_ioc_transfer *t,
                                 uint8_t * buffer,
                                 uint8_t   columnMask)
    {
      mcp23s17FillSetRegisterTransfer(t,
                                  buffer,
                                  7,
                                  Mcp23s17Reg::GPIOB,
                                  columnMask);
      t->cs_change = 0;
    }
};


class SN74594OutputTransfer : public SpiTransfer
{
private:
  uint8_t   m_columnBuffer;
  struct spi_ioc_transfer m_transfers[2];

public:
  SN74594OutputTransfer (int column,
                         uint8_t *ledBuffer
                          )
  {
    memset(m_transfers, 0, sizeof(m_transfers));

    m_columnBuffer = 1<<column;
    {
      struct spi_ioc_transfer *t0 = &m_transfers[0];
      t0->tx_buf = (unsigned long)&m_columnBuffer;
      t0->rx_buf = 0;
      t0->len = 1;
      t0->delay_usecs = 0;
      t0->speed_hz = 0;
      t0->bits_per_word = 8;
      t0->cs_change = 0;
    }
    {
      struct spi_ioc_transfer *t1 = &m_transfers[1];
      t1->tx_buf = (unsigned long)ledBuffer;
      t1->rx_buf = 0;
      t1->len = 8;
      t1->delay_usecs = 0;
      t1->speed_hz = 0;
      t1->bits_per_word = 8;
      t1->cs_change = 0;
    }
  }

  int count() const
  {
    return sizeof(m_transfers)/sizeof(m_transfers[0]);
  }

  const struct spi_ioc_transfer * transfers() const
  {
    return m_transfers;
  }
};





class TestOutputTransfer : public SpiTransfer
{
private:
  uint8_t   m_testBuffer[3];
  struct spi_ioc_transfer m_transfers[3];

public:
  TestOutputTransfer (int column,
                         uint8_t *ledBuffer
                          )
  {
    memset(m_transfers, 0, sizeof(m_transfers));

    m_testBuffer[0] = 0xff;
    m_testBuffer[1] = 0xff;
    m_testBuffer[2] = 0xff;

    {
      struct spi_ioc_transfer *t0 = &m_transfers[0];
      t0->tx_buf = (unsigned long)&m_testBuffer[0];
      t0->rx_buf = 0;
      t0->len = 1;
      t0->delay_usecs = 0;
      t0->speed_hz = 0;
      t0->bits_per_word = 8;
      t0->cs_change = 1;
    }
    {
      struct spi_ioc_transfer *t0 = &m_transfers[1];
      t0->tx_buf = (unsigned long)&m_testBuffer[1];
      t0->rx_buf = 0;
      t0->len = 1;
      t0->delay_usecs = 0;
      t0->speed_hz = 0;
      t0->bits_per_word = 8;
      t0->cs_change = 1;
    }
    {
      struct spi_ioc_transfer *t0 = &m_transfers[2];
      t0->tx_buf = (unsigned long)&m_testBuffer[2];
      t0->rx_buf = 0;
      t0->len = 1;
      t0->delay_usecs = 0;
      t0->speed_hz = 0;
      t0->bits_per_word = 8;
      t0->cs_change = 0;
    }
  }

  int count() const
  {
    return sizeof(m_transfers)/sizeof(m_transfers[0]);
  }

  const struct spi_ioc_transfer * transfers() const
  {
    return m_transfers;
  }
};



class LedCube8x8x8
{
private:
  enum { LayerCount = 8 };
  std::vector<SpiTransfer*> m_spiTransfers;
  int      spi_fd = -1;
  uint8_t  m_ledBits[512/8];

public:
  ~LedCube8x8x8()
  {
    for (auto i : m_spiTransfers)
      delete i;
  }

  void Initialize(OutputTransferFactory * f)
  {
    if (spi_fd < 0)
      throw std::runtime_error("OpenSpi must be called before Initialize");

    SpiTransfer * t = f->createConfigureSpiTransfer();
    if (t)
      {
        std::cout << "initialize output" << std::endl;
        StartSpiTransfer(spi_fd, t);
        delete t;
      }

    for (auto i = 0; i < LayerCount; ++i)
      m_spiTransfers.push_back(
                               f->createOutputSpiTransfer(i, &m_ledBits[8*i])
                               );
  }

  void OpenSpi(const char * spiDevice,
               uint8_t  mode,
               uint32_t speed)
  {
    spi_fd = ::open(spiDevice, O_RDWR);
    if (spi_fd < 0)
      throw std::runtime_error(std::string("failed to open spi device ")+spiDevice);

    /*
     * spi mode
     */
    int ret = ioctl(spi_fd, SPI_IOC_WR_MODE, &mode);
    if (ret == -1)
      throw std::runtime_error("can't set spi mode");
    
    ret = ioctl(spi_fd, SPI_IOC_RD_MODE, &mode);
    if (ret == -1)
      throw std::runtime_error("can't get spi mode");

    /*
     * bits per word
     */
    {
      int bits = 8;
      ret = ioctl(spi_fd, SPI_IOC_WR_BITS_PER_WORD, &bits);
      if (ret == -1)
        throw std::runtime_error("can't set bits per word");
      
      ret = ioctl(spi_fd, SPI_IOC_RD_BITS_PER_WORD, &bits);
      if (ret == -1)
        throw std::runtime_error("can't get bits per word");
      std::cout << "bits per word: " << bits << std::endl;
    }

    /*
     * max speed hz
     */
    ret = ioctl(spi_fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
    if (ret == -1)
      throw std::runtime_error("can't set max speed hz");
    
    ret = ioctl(spi_fd, SPI_IOC_RD_MAX_SPEED_HZ, &speed);
    if (ret == -1)
      throw std::runtime_error("can't get max speed hz");

    std::cout << "spi mode: " << mode << std::endl;
    std::cout << "max speed: " << speed << " Hz (" << (speed / 1000) << " KHz)" << std::endl;
  }

  void setChannel(int channel, int value)
  {
    if (channel < 512)
      {
        if (value >= 0x80)
          m_ledBits[channel/8] |= (1<<(channel % 8));
        else
          m_ledBits[channel/8] &= ~(1<<(channel % 8));
      }
  }

  void OutputLoop(std::function<bool(void)> run)
  {
    std::chrono::system_clock::time_point nextTransferStartTime =
      std::chrono::system_clock::now();

    while (run())
      {
        for (auto transfer : m_spiTransfers)
          {
            std::this_thread::sleep_until(nextTransferStartTime);
            StartSpiTransfer(spi_fd, transfer);
            nextTransferStartTime += std::chrono::microseconds(long(1000));
          }
      }
  }

  void StartSpiTransfer(int fd, SpiTransfer*t)
  {
    ioctl(fd, SPI_IOC_MESSAGE(t->count()), t->transfers());
  }
};

static std::string  spiType = "mcp23s17";
static const char *spi_device = "/dev/spidev0.0";
static uint8_t  mode = 0;
static uint32_t speed = 5*1000*1000; // 5MHz, the mcp23s17 supports up to 10MHz.
static uint16_t delay = 0;
static int      frameCount = 10000;

static void print_usage(const char *prog);
static void parse_opts(int argc, char *argv[])
{
	while (1) {
		static const struct option lopts[] = {
			{ "device",  1, 0, 'D' },
			{ "speed",   1, 0, 's' },
			{ "delay",   1, 0, 'd' },
			{ "bpw",     1, 0, 'b' },
			{ "loop",    0, 0, 'l' },
			{ "cpha",    0, 0, 'H' },
			{ "cpol",    0, 0, 'O' },
			{ "lsb",     0, 0, 'L' },
			{ "cs-high", 0, 0, 'C' },
			{ "3wire",   0, 0, '3' },
			{ "no-cs",   0, 0, 'N' },
			{ "ready",   0, 0, 'R' },
			{ "spitype", 1, 0, 'T' },
                        
			{ NULL, 0, 0, 0 },
		};
		int c;

		c = getopt_long(argc, argv, "D:s:d:b:lHOLC3NRT:F:", lopts, NULL);

		if (c == -1)
			break;

		switch (c) {
		case 'D':
			spi_device = optarg;
			break;
		case 's':
			speed = atoi(optarg);
			break;
		case 'd':
			delay = atoi(optarg);
			break;
		case 'l':
			mode |= SPI_LOOP;
			break;
		case 'H':
			mode |= SPI_CPHA;
			break;
		case 'O':
			mode |= SPI_CPOL;
			break;
		case 'L':
			mode |= SPI_LSB_FIRST;
			break;
		case 'C':
			mode |= SPI_CS_HIGH;
			break;
		case '3':
			mode |= SPI_3WIRE;
			break;
		case 'N':
			mode |= SPI_NO_CS;
			break;
		case 'R':
			mode |= SPI_READY;
			break;
                case 'T':
                        spiType =  std::string(optarg);
                        break;
                case 'F':
                        frameCount = atoi(optarg);
                        break;
		default:
			print_usage(argv[0]);
			break;
		}
	}
}


class TestTransferFactory : public OutputTransferFactory
{
public:
  SpiTransfer * createOutputSpiTransfer(int i, uint8_t* buf) const
  {
    return new TestOutputTransfer (i, buf);
  }

  SpiTransfer * createConfigureSpiTransfer() const
  {
    return 0;
  }
};

class Mcp23S17TransferFactory : public OutputTransferFactory
{
public:
  SpiTransfer * createOutputSpiTransfer(int i, uint8_t* buf) const
  {
    return new Mcp23S17OutputTransfer (i, buf);
  }

  SpiTransfer * createConfigureSpiTransfer() const
  {
    return new Mcp23S17ConfigureTransfer;
  }
};

class SN74594TransferFactory : public OutputTransferFactory
{
public:
  SpiTransfer * createOutputSpiTransfer(int i, uint8_t* buf) const
  {
    return new SN74594OutputTransfer (i, buf);
  }

  SpiTransfer * createConfigureSpiTransfer() const
  {
    return 0;
  }
};


static const std::map<std::string, std::function<OutputTransferFactory*()> > spiTypeCreateFunctions
{
  { "Test",
      []() -> OutputTransferFactory* { return new TestTransferFactory; }
  },
    { "mcp23s17",
        []() -> OutputTransferFactory*   { return new Mcp23S17TransferFactory; }
    },
    { "sn74594",
        []() -> OutputTransferFactory*   { return new SN74594TransferFactory; }
    }
};


OutputTransferFactory * spiTransferFactory (const std::string & t)
{
  auto i = spiTypeCreateFunctions.find(t);
  if (i == spiTypeCreateFunctions.end())
    throw std::invalid_argument("unknown spi transfer type");
  return i->second();
}


static void print_usage(const char *prog)
{
  std::cout << "Usage: " << prog << " [-DsbdlHOLC3]" << std::endl;
  std::cout << "  -D --device   device to use (default /dev/spidev1.1)" << std::endl
            << "  -s --speed    max speed (Hz)" << std::endl
            << "  -d --delay    delay (usec)" << std::endl
            << "  -l --loop     loopback" << std::endl
            << "  -H --cpha     clock phase" << std::endl
            << "  -O --cpol     clock polarity" << std::endl
            << "  -L --lsb      least significant bit first" << std::endl
            << "  -C --cs-high  chip select active high" << std::endl
            << "  -3 --3wire    SI/SO signals shared" << std::endl
            << "  -F            number of frames to output" << std::endl
            << "  -T --spitype  output functions:";
  for (auto i : spiTypeCreateFunctions)
    std::cout << " " << i.first;
  std::cout << std::endl;
  exit(1);
}


#include <sys/mman.h>
namespace Scheduling
{
  enum Policy {RoundRobin = SCHED_RR};

  /* Alternative to `chrt -r -p 99 <pid>`. */
  inline void setPriority(double priority = 1.0, Policy policy = RoundRobin)
  {
    if (priority<0 || priority>1.0)
      throw std::invalid_argument(std::string("setPriority: priority out of range"));

    /* Lock all current and future pages from preventing of being paged to swap */
    if (mlockall( MCL_CURRENT | MCL_FUTURE ))
        throw std::runtime_error(std::string("failed to lock memory for realtime: ") + strerror(errno));
    
    int min_priority = sched_get_priority_min(policy), max_priority = sched_get_priority_max(policy);
    int adjusted_priority = int((max_priority-min_priority) * priority) + min_priority;
    sched_param param = {adjusted_priority};
    if (sched_setscheduler(0, policy, &param))
      throw std::runtime_error(std::string("failed to set thread priority."));
  }
}

int main (int argc, char **argv)
{
  LedCube8x8x8  cube;

  parse_opts(argc, argv);

  for (auto i = 0; i < 512; ++i)
    cube.setChannel(i, i%256);

  try
    {
      cube.OpenSpi(spi_device, mode, speed);
    }
  catch(std::exception e)
   {
     std::cout << "failed to open spi:" << e.what() << std::endl;
     return 1;
   }

  auto spiCreator = spiTransferFactory (spiType);
  if (spiCreator)
    cube.Initialize(spiCreator);

  std::this_thread::sleep_for(std::chrono::milliseconds(long(10)));

  int framesRemaining = frameCount;
  std::function<bool(void)> checkrun = [&framesRemaining]()->bool
    {
      return framesRemaining-- > 0;
    };

  bool threadded = true;
  if (threadded)
    {
      std::thread loopThread([&cube,&checkrun]()
                             {
                               Scheduling::setPriority(1.0, Scheduling::RoundRobin);
                               cube.OutputLoop(checkrun);
                             });

      while (framesRemaining > 0)
        for (auto k = 0; k < 100; ++k)
          {
            std::cout << k << std::endl;
            for (auto i = 0; i < 512; ++i)
              cube.setChannel(i, (k+i)%256);
            std::this_thread::sleep_for(std::chrono::milliseconds(long(100)));
          }

      loopThread.join();
    }
  else
    cube.OutputLoop(checkrun);

  return 0;
}

